import os
from typing import Any, Callable, List, Optional, Tuple

import numpy as np
import pandas as pd
from dask_ml.compose import make_column_transformer
from dask_ml.preprocessing import Categorizer, OrdinalEncoder
from git_root import git_root
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer

from ..util import exposure_and_offset_to_weights

# taken from https://github.com/lorentzenchr/Tutorial_freMTPL2/blob/master/glm_freMTPL2_example.ipynb  # noqa: B950
# Modified to generate data sets of different sizes


def create_insurance_raw_data(verbose=False) -> None:
    """Do some basic processing on the data that we will later transform into our \
    benchmark data sets."""
    # load the datasets
    # first row (=column names) uses "", all other rows use ''
    # use '' as quotechar as it is easier to change column names
    df = pd.read_csv(
        "https://www.openml.org/data/get_csv/20649148/freMTPL2freq.arff", quotechar="'"
    )

    # rename column names '"name"' => 'name'
    df.rename(lambda x: x.replace('"', ""), axis="columns", inplace=True)
    df["IDpol"] = df["IDpol"].astype(np.int64)
    df.set_index("IDpol", inplace=True)

    df_sev = pd.read_csv(
        "https://www.openml.org/data/get_csv/20649149/freMTPL2sev.arff", index_col=0
    )

    # join ClaimAmount from df_sev to df:
    #   1. cut ClaimAmount at 100_000
    #   2. aggregate ClaimAmount per IDpol
    #   3. join by IDpol
    df_sev["ClaimAmountCut"] = df_sev["ClaimAmount"].clip(upper=100_000)
    df = df.join(df_sev.groupby(level=0).sum(), how="left")
    df.fillna(value={"ClaimAmount": 0, "ClaimAmountCut": 0}, inplace=True)

    # Check if there are IDpol in df_sev that do not match any IDPol in df.
    df2 = pd.merge(
        df_sev,
        df.loc[:, ["ClaimNb"]],
        left_index=True,
        right_index=True,
        how="outer",
        indicator=True,
    )
    if verbose:
        print(
            "There are {} rows in freMTPL2sev that do not have a matching IDpol in "
            "freMTPL2freq. They have a ClaimAmountCut of {}.".format(
                df2[df2._merge == "left_only"].shape[0],
                df2.ClaimAmountCut[df2._merge == "left_only"].sum(),
            )
        )

    round(df_sev.ClaimAmountCut.sum() - df.ClaimAmountCut.sum(), 2)

    if verbose:
        print(
            "Number or rows with ClaimAmountCut > 0 and ClaimNb == 0: {}".format(
                df[(df.ClaimAmountCut > 0) & (df.ClaimNb == 0)].shape[0]
            )
        )

    # 9116 zero claims
    if verbose:
        print(
            "Number or rows with ClaimAmountCut = 0 and ClaimNb >= 1: {}".format(
                df[(df.ClaimAmountCut == 0) & (df.ClaimNb >= 1)].shape[0]
            )
        )

    # Note: Zero claims must be ignored in severity models, because the support is
    # (0, inf) not [0, inf). Therefore, we define the number of claims with positive
    # claim amount for later use.
    df["ClaimNb_pos"] = df["ClaimNb"]
    df.loc[(df.ClaimAmount <= 0) & (df.ClaimNb >= 1), "ClaimNb_pos"] = 0

    # correct for unreasonable observations (that might be data error)
    # see case study paper
    df["ClaimNb"] = df["ClaimNb"].clip(upper=4)
    df["ClaimNb_pos"] = df["ClaimNb_pos"].clip(upper=4)
    df["Exposure"] = df["Exposure"].clip(upper=1)

    out_path = git_root("data/insurance.parquet")
    os.makedirs(os.path.dirname(out_path), exist_ok=True)
    df.to_parquet(out_path)


def get_categorizer(col_name: str, name="cat") -> Tuple[str, Categorizer]:
    """
    Get a dask_ml Categorizer.

    Categorizer only operates on object columns unless you explictily pass the column
    name.
    """
    return name, Categorizer(columns=[col_name])


def func_returns_df(
    fn: Callable[[pd.DataFrame], np.ndarray]
) -> Callable[[pd.DataFrame], pd.DataFrame]:
    """
    Take a function that takes a dataframe and returns a Numpy array, and return a \
    function that takes a dataframe and returns a dataframe.

    fn: Function that takes a dataframe and returns a numpy array
    Returns: Function that takes a dataframe and returns a dataframe with the values
             determined by the original function, and the index and columns of the original
             dataframe.
    """
    return lambda x: x.assign(**{x.columns[0]: fn(x)})


def gen_col_trans() -> Tuple[Any, List[str]]:
    """Generate a ColumnTransformer and list of names.

    The transformer corresponds to the GLM of the case study paper.

    Encodes k categories with k binary features (redundant).
    """
    column_trans = make_column_transformer(
        # VehPower 4, 5, 6, 7, 8, 9, drop=4
        (
            Pipeline(
                [
                    (
                        "cut_9",
                        FunctionTransformer(lambda x: np.minimum(x, 9), validate=False),
                    ),
                    get_categorizer("VehPower"),
                ]
            ),
            ["VehPower"],
        ),
        # VehAge intervals [0,1), [1, 10], (10, inf), drop=[1,10]
        (
            Pipeline(
                [
                    (
                        "bin",
                        FunctionTransformer(
                            func_returns_df(
                                lambda x: np.digitize(
                                    np.where(x == 10, 9, x), bins=[1, 10]
                                )
                            ),
                            validate=False,
                        ),
                    ),
                    get_categorizer("VehAge"),
                ]
            ),
            ["VehAge"],
        ),
        # DrivAge intervals [18,21), [21,26), [26,31), [31,41), [41,51), [51,71),[71,∞),
        # drop=[41,51)
        (
            Pipeline(
                [
                    (
                        "bin",
                        FunctionTransformer(
                            func_returns_df(
                                lambda x: np.digitize(x, bins=[21, 26, 31, 41, 51, 71])
                            ),
                            validate=False,
                        ),
                    ),
                    get_categorizer("DrivAge"),
                ]
            ),
            ["DrivAge"],
        ),
        (
            Pipeline(
                [
                    (
                        "cutat150",
                        FunctionTransformer(
                            lambda x: np.minimum(x, 150), validate=False
                        ),
                    )
                ]
            ),
            ["BonusMalus"],
        ),
        (Pipeline([get_categorizer("VehBrand")]), ["VehBrand"]),
        (Pipeline([get_categorizer("VehGas")]), ["VehGas"]),
        (FunctionTransformer(np.log, validate=False), ["Density"]),
        (Pipeline([get_categorizer("Region")]), ["Region"]),
        (
            Pipeline(
                [
                    get_categorizer("Area"),
                    ("OE", OrdinalEncoder()),
                    (
                        "plus_1",
                        FunctionTransformer(lambda x: x + 1, validate=False),
                    ),
                ]
            ),
            ["Area"],
        ),
        remainder="drop",
    )
    column_trans_names = [
        "VehPower_4",
        "VehPower_5",
        "VehPower_6",
        "VehPower_7",
        "VehPower_8",
        "VehPower_9",
        "VehAge_[0,1)",
        "VehAge_[1, 10]",
        "VehAge_(10,inf)",
        "DrivAge_[18,21)",
        "DrivAge_[21,26)",
        "DrivAge_[26,31)",
        "DrivAge_[31,41)",
        "DrivAge_[41,51)",
        "DrivAge_[51,71)",
        "DrivAge_[71,inf)",
        "BonusMalus",
        "VehBrand_B10",
        "VehBrand_B11",
        "VehBrand_B12",
        "VehBrand_B13",
        "VehBrand_B14",
        "VehBrand_B1",
        "VehBrand_B2",
        "VehBrand_B3",
        "VehBrand_B4",
        "VehBrand_B5",
        "VehBrand_B6",
        "VehGas_Diesel",
        "VehGas_Regular",
        "Density_log",
        "Region_R11",
        "Region_R21",
        "Region_R22",
        "Region_R23",
        "Region_R24",
        "Region_R25",
        "Region_R26",
        "Region_R31",
        "Region_R41",
        "Region_R42",
        "Region_R43",
        "Region_R52",
        "Region_R53",
        "Region_R54",
        "Region_R72",
        "Region_R73",
        "Region_R74",
        "Region_R82",
        "Region_R83",
        "Region_R91",
        "Region_R93",
        "Region_R94",
        "Area_ord",
    ]
    column_trans_names = [
        i
        for i in column_trans_names
        if i
        not in [
            "VehPower_4",
            "VehAge_[1, 10]",
            "DrivAge_[41,51)",
            "VehBrand_B1",
            "VehGas_Diesel",
            "Region_R24",
        ]
    ]
    return column_trans, column_trans_names


def add_noise(df: pd.DataFrame, noise: float) -> pd.DataFrame:
    """Add noise by swapping out data points."""
    np.random.seed(43212)
    for col in df.columns:
        if col in ["ClaimNb", "Exposure", "ClaimAmountCut", "ClaimNb_pos"]:
            continue
        swap = np.random.uniform(size=len(df)) < noise
        shuffle = np.random.choice(df[col], size=len(df))
        df.loc[swap, col] = shuffle[swap]

    return df


def compute_y_exposure(df, distribution):
    """Compute y and exposure depending on distribution."""
    if distribution == "poisson":
        exposure = df["Exposure"].values
        y = df["ClaimNb"].values / exposure
    elif distribution in ["gamma", "gaussian"]:
        exposure = df["ClaimNb_pos"].values
        y = df["ClaimAmountCut"].values / exposure
    elif "tweedie" in distribution:
        exposure = df["Exposure"].values
        y = df["ClaimAmountCut"].values / exposure
    elif distribution == "binomial":
        exposure = df["Exposure"].values
        df["HasClaim"] = (df["ClaimNb"] > 0).astype(np.int32)
        y = df["HasClaim"].values
    else:
        raise ValueError(
            "distribution must be one of ['poisson', 'gamma', 'tweedie', 'gaussian', "
            f"'binomial'] not {distribution}."
        )

    return y, exposure


def _read_insurance_data(
    num_rows: Optional[int], noise: Optional[float], distribution: str
) -> pd.DataFrame:
    path = git_root("data/insurance.parquet")
    if not os.path.exists(path):
        create_insurance_raw_data()
    df = pd.read_parquet(path)

    if distribution in ["gamma", "gaussian"]:
        df = df.query("ClaimAmountCut > 0")

    if num_rows is not None:
        # if we're oversampling, set default value for noise to 0.05
        # can be turned off by setting noise to zero
        if noise is None and num_rows > len(df):
            noise = 0.05
        df = df.sample(n=num_rows, replace=True, random_state=12345)

    if noise is not None:
        df = add_noise(df, noise=noise)
    return df


def generate_narrow_insurance_dataset(
    num_rows=None, noise=None, distribution="poisson"
) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray]:
    """Generate the tutorial data set from the sklearn fork and save it to disk."""
    df = _read_insurance_data(num_rows, noise, distribution)

    col_trans_GLM1, _ = gen_col_trans()
    y, exposure = compute_y_exposure(df, distribution)

    return col_trans_GLM1.fit_transform(df), y, exposure


def generate_real_insurance_dataset(
    num_rows=None, noise=None, distribution="poisson"
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Load real insurance data set."""
    df = pd.read_parquet(git_root("data", "outcomes.parquet"))
    X = pd.read_parquet(git_root("data", "X.parquet"))

    if distribution != "poisson":
        raise NotImplementedError("distribution must be poisson")

    # restrict X and df to train set
    train_set = df["sample"] == "train"
    X = X.loc[train_set].reset_index(drop=True)
    df = df.loc[train_set].reset_index(drop=True)

    # subsample
    if num_rows is not None:
        idx = df.sample(n=num_rows).index
        df = df.loc[idx].reset_index(drop=True)
        X = X.loc[idx].reset_index(drop=True)

    # account for exposure and offsets
    y, weights = exposure_and_offset_to_weights(
        power=1,
        y=df["sanzkh02"],
        exposure=df["je"],
        offset=df["offset_kh_sach_frequenz"],
    )

    return X, y, weights


def generate_wide_insurance_dataset(
    num_rows=None, noise=None, distribution="poisson"
) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray]:
    """Generate a version of the tutorial data set with many features."""
    df = _read_insurance_data(num_rows, noise, distribution)
    cat_cols = [
        "Area",
        "VehPower",
        "VehAge",
        "DrivAge",
        "BonusMalus",
        "VehBrand",
        "VehGas",
        "Region",
    ]

    transformer = make_column_transformer(
        (
            FunctionTransformer(),
            lambda x: [
                elmt
                for elmt in x.select_dtypes(["number"]).columns
                if elmt not in cat_cols
            ],
        ),
        (
            Pipeline([get_categorizer(col, "cat_" + col) for col in cat_cols]),
            cat_cols,
        ),
        remainder="drop",
    )
    y, exposure = compute_y_exposure(df, distribution)

    return transformer.fit_transform(df), y, exposure


def generate_intermediate_insurance_dataset(
    num_rows=None, noise=None, distribution="poisson"
) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray]:
    """Generate the tutorial data set from the sklearn fork and save it to disk."""
    df = _read_insurance_data(num_rows, noise, distribution)
    df["BonusMalusClipped"] = df["BonusMalus"].clip(50, 100)

    col_trans_GLM1, _ = gen_col_trans()
    col_trans_GLM1.transformers.append(
        (
            "BonusMalusClipped",
            Pipeline([get_categorizer("BonusMalusClipped")]),
            ["BonusMalusClipped"],
        )
    )
    y, exposure = compute_y_exposure(df, distribution)

    return col_trans_GLM1.fit_transform(df), y, exposure



# cython: boundscheck=False, wraparound=False, cdivision=True

from cython cimport floating
from cython.parallel import prange

from libc.math cimport M_PI, ceil, exp, floor, fmax, lgamma, log

import numpy as np

ctypedef fused numeric:
    short
    int
    long
    float
    double

# If an argument is readonly, that will fail with a typical floating[:]
# memoryview. However, const floating[:] causes failures because currently,
# const and fused types (floating) and memoryviews cannot all be combined in
# Cython. https://github.com/cython/cython/issues/1772
cdef fused const_floating1d:
    const float[:]
    const double[:]

# NOTE: Here and below, the factor argument is left for last. That's because it
# will be a single floating point value being passed from Python. In Python,
# all floating point values are 64-bit. As a result, if it's the first
# argument, it will cause the fused types to always evaluate to double.
# Whereas, if it's the last argument, by the time it's encountered, floating
# will have already been set and the 64-bit float will be cast to the correct
# time.
def normal_identity_eta_mu_deviance(
    const_floating1d cur_eta,
    const_floating1d X_dot_d,
    const_floating1d y,
    const_floating1d weights,
    floating[:] eta_out,
    floating[:] mu_out,
    floating factor
):
    cdef int n = cur_eta.shape[0]
    cdef int i
    cdef floating unit_deviance
    cdef floating deviance = 0.0
    for i in prange(n, nogil=True):
        eta_out[i] = cur_eta[i] + factor * X_dot_d[i]
        mu_out[i] = eta_out[i]
        # Note: deviance is equal to -2 times the true log likelihood to match
        # the default calculation using unit_deviance in _distribution.py
        # True log likelihood: -1/2 * (y - mu)**2
        deviance += weights[i] * (y[i] - mu_out[i]) ** 2
    return deviance

def normal_identity_rowwise_gradient_hessian(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d eta,
    const_floating1d mu,
    floating[:] gradient_rows_out,
    floating[:] hessian_rows_out
):
    cdef int n = eta.shape[0]
    cdef int i
    for i in prange(n, nogil=True):
        gradient_rows_out[i] = weights[i] * (y[i] - mu[i])
        # Note: hessian_rows_out yields -1 times the true hessian to match
        # the default calculation in _distribution.py
        hessian_rows_out[i] = weights[i]

def normal_log_likelihood(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating dispersion,
):
    cdef int i  # loop counter
    cdef floating sum_weights  # helper

    cdef int n = y.shape[0]  # loop length
    cdef floating ll = 0.0  # output

    for i in prange(n, nogil=True):
        ll -= weights[i] * (y[i] - mu[i]) ** 2
        sum_weights -= weights[i]

    return ll / (2 * dispersion) + sum_weights * log(2 * M_PI * dispersion) / 2

def normal_deviance(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating dispersion,
):
    cdef int i  # loop counter

    cdef int n = y.shape[0]  # loop length
    cdef floating D = 0.0  # output

    for i in prange(n, nogil=True):
        D += weights[i] * (y[i] - mu[i]) ** 2

    return D

def poisson_log_eta_mu_deviance(
    const_floating1d cur_eta,
    const_floating1d X_dot_d,
    const_floating1d y,
    const_floating1d weights,
    floating[:] eta_out,
    floating[:] mu_out,
    floating factor
):
    cdef int n = cur_eta.shape[0]
    cdef int i
    cdef floating unit_deviance
    cdef floating deviance = 0.0
    for i in prange(n, nogil=True):
        eta_out[i] = cur_eta[i] + factor * X_dot_d[i]
        mu_out[i] = exp(eta_out[i])
        # True log likelihood: y * eta - mu - lgamma(1 + y)
        deviance += weights[i] * (y[i] * eta_out[i] - mu_out[i])
    return -2 * deviance

def poisson_log_rowwise_gradient_hessian(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d eta,
    const_floating1d mu,
    floating[:] gradient_rows_out,
    floating[:] hessian_rows_out
):
    cdef int n = eta.shape[0]
    cdef int i
    for i in prange(n, nogil=True):
        gradient_rows_out[i] = weights[i] * (y[i] - mu[i])
        hessian_rows_out[i] = weights[i] * mu[i]

def poisson_log_likelihood(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating dispersion,
):
    cdef int i  # loop counter

    cdef int n = y.shape[0]  # loop length
    cdef floating ll = 0.0  # output

    for i in prange(n, nogil=True):
        ll -= weights[i] * mu[i]
        if y[i] > 0:
            ll -= weights[i] * (lgamma(1 + y[i]) - y[i] * log(mu[i]))

    return ll

def poisson_deviance(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating dispersion,
):
    cdef int i  # loop counter

    cdef int n = y.shape[0]  # loop length
    cdef floating D = 0.0  # output

    for i in prange(n, nogil=True):
        D += weights[i] * mu[i]
        if y[i] > 0:
            D += weights[i] * y[i] * (log(y[i]) - log(mu[i]) - 1)

    return 2 * D

def gamma_log_eta_mu_deviance(
    const_floating1d cur_eta,
    const_floating1d X_dot_d,
    const_floating1d y,
    const_floating1d weights,
    floating[:] eta_out,
    floating[:] mu_out,
    floating factor
):
    cdef int n = cur_eta.shape[0]
    cdef int i
    cdef floating unit_deviance
    cdef floating deviance = 0.0
    for i in prange(n, nogil=True):
        eta_out[i] = cur_eta[i] + factor * X_dot_d[i]
        mu_out[i] = exp(eta_out[i])
        # True log likelihood: -(y / mu + eta)
        deviance += weights[i] * (y[i] / mu_out[i] + eta_out[i])
    return 2 * deviance

def gamma_log_rowwise_gradient_hessian(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d eta,
    const_floating1d mu,
    floating[:] gradient_rows_out,
    floating[:] hessian_rows_out
):
    cdef int n = eta.shape[0]
    cdef int i
    for i in prange(n, nogil=True):
        gradient_rows_out[i] = weights[i] * (y[i] / mu[i] - 1)
        hessian_rows_out[i] = weights[i] * (y[i] / mu[i])

def gamma_log_likelihood(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating dispersion,
):
    cdef int i  # loop counter
    cdef floating ln_y, sum_weights  # helpers

    cdef int n = y.shape[0]  # loop length
    cdef floating ll = 0.0  # output
    cdef floating inv_dispersion = 1 / dispersion
    cdef floating normalization = log(dispersion) * inv_dispersion + lgamma(inv_dispersion)

    for i in prange(n, nogil=True):
        ln_y = log(y[i])
        ll += weights[i] * (inv_dispersion * (ln_y - log(mu[i]) - y[i] / mu[i]) - ln_y)
        sum_weights += weights[i]

    return ll - normalization * sum_weights

def gamma_deviance(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating dispersion,
):
    cdef int i  # loop counter

    cdef int n = y.shape[0]  # loop length
    cdef floating D = 0.0  # output

    for i in prange(n, nogil=True):
        D += weights[i] * (log(mu[i]) - log(y[i]) + y[i] / mu[i] - 1)

    return 2 * D

def tweedie_log_eta_mu_deviance(
    const_floating1d cur_eta,
    const_floating1d X_dot_d,
    const_floating1d y,
    const_floating1d weights,
    floating[:] eta_out,
    floating[:] mu_out,
    floating factor,
    floating p
):
    cdef int n = cur_eta.shape[0]
    cdef int i
    cdef floating unit_deviance
    cdef floating deviance = 0.0
    cdef floating mu1mp
    for i in prange(n, nogil=True):
        eta_out[i] = cur_eta[i] + factor * X_dot_d[i]
        mu_out[i] = exp(eta_out[i])
        mu1mp = mu_out[i] ** (1 - p)
        deviance += weights[i] * mu1mp * (mu_out[i] / (2 - p) - y[i] / (1 - p))
    return 2 * deviance

def tweedie_log_rowwise_gradient_hessian(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d eta,
    const_floating1d mu,
    floating[:] gradient_rows_out,
    floating[:] hessian_rows_out,
    floating p
):
    cdef int n = eta.shape[0]
    cdef int i
    cdef floating mu1mp, ymm
    for i in prange(n, nogil=True):
        mu1mp = mu[i] ** (1 - p)
        ymm = y[i] - mu[i]
        gradient_rows_out[i] = weights[i] * mu1mp * ymm
        # This hessian will be positive definite for 1 < p < 2. Don't use it
        # outside that range.
        hessian_rows_out[i] = weights[i] * mu1mp * (mu[i] - (1 - p) * ymm)

def tweedie_deviance(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating p,
):
    cdef int i  # loop counter
    cdef floating mu1mp, yo1mp  # helpers

    cdef int n = y.shape[0]  # loop length
    cdef floating D = 0.0  # output

    for i in prange(n, nogil=True):

        mu1mp = mu[i] ** (1 - p)
        D += weights[i] * (mu1mp * mu[i]) / (2 - p)

        if y[i] > 0:
            yo1mp = y[i] / (1 - p)
            D += weights[i] * ((y[i] ** (1 - p)) * yo1mp / (2 - p) - yo1mp * mu1mp)

    return 2 * D

def tweedie_log_likelihood(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d mu,
    floating p,
    floating dispersion,
):
    cdef int i  # loop counter

    cdef int n = y.shape[0]  # loop length
    cdef floating ll = 0.0  # output

    for i in prange(n, nogil=True):
        ll += weights[i] * _tweedie_unit_loglikelihood(y[i], mu[i], p, dispersion)

    return ll

cdef floating _tweedie_unit_loglikelihood(floating y, floating mu, floating power, floating dispersion) nogil:

    cdef floating kappa, normalization, theta

    if y == 0:
        return -(mu ** (2 - power)) / (dispersion * (2 - power))
    else:
        theta = mu ** (1 - power)
        kappa = mu * theta / (2 - power)
        theta = theta / (1 - power)
        normalization = _tweedie_normalization(y, power, dispersion)
        return (theta * y - kappa) / dispersion + normalization

cdef floating _tweedie_normalization(floating y, floating power, floating dispersion) nogil:

    cdef int j, j_lower, j_upper

    cdef floating j_max = exp((2 - power) * log(y) - log(dispersion) - log(2 - power))
    cdef floating w_max = _log_w_j(y, power, dispersion, j_max)
    cdef floating w_summand = 0.0

    j_lower, j_upper = _sum_limits(y, power, dispersion, j_max)

    for j in range(j_lower, j_upper + 1):
        w_summand += exp(_log_w_j(y, power, dispersion, j) - w_max)

    return w_max + log(w_summand) - log(y)

cdef (int, int) _sum_limits(floating y, floating power, floating dispersion, floating j_max) nogil:

    cdef floating w_lower

    cdef floating j_lower = 1.0
    cdef floating j_upper = ceil(j_max)
    cdef floating w_upper = _log_w_j(y, power, dispersion, j_upper)

    cdef floating w_crt = _log_w_j(y, power, dispersion, j_max) - 37
    cdef floating w_one = _log_w_j(y, power, dispersion, 1)

    if w_one <= w_crt:
        j_lower = floor(j_max)
        w_lower = _log_w_j(y, power, dispersion, j_lower)
        while w_lower >= w_crt:
            j_lower -= 1
            w_lower = _log_w_j(y, power, dispersion, j_lower)
    while w_upper >= w_crt:
        j_upper += 1
        w_upper = _log_w_j(y, power, dispersion, j_upper)

    return int(j_lower), int(j_upper)

cdef floating _log_w_j(floating y, floating power, floating dispersion, numeric j) nogil:
    cdef floating alpha = (2 - power) / (1 - power)
    return j * _log_z(y, power, dispersion) - lgamma(1 + j) - lgamma(-alpha * j)

cdef floating _log_z(floating y, floating power, floating dispersion) nogil:
    cdef floating alpha = (2 - power) / (1 - power)
    return (
        alpha * log(power - 1)
        - alpha * log(y)
        - (1 - alpha) * log(dispersion)
        - log(2 - power)
    )

def binomial_logit_eta_mu_deviance(
    const_floating1d cur_eta,
    const_floating1d X_dot_d,
    const_floating1d y,
    const_floating1d weights,
    floating[:] eta_out,
    floating[:] mu_out,
    floating factor
):
    cdef int n = cur_eta.shape[0]
    cdef int i
    cdef floating unit_deviance
    cdef floating deviance = 0.0
    cdef floating expposeta, expnegeta
    for i in prange(n, nogil=True):
        eta_out[i] = cur_eta[i] + factor * X_dot_d[i]
        # When eta is positive, we want to use formulas that depend on
        # exp(-eta), but when eta is negative we want to use formulas that
        # depend on exp(eta), rederived based on the suggestions here:
        # http://fa.bianp.net/blog/2013/numerical-optimizers-for-logistic-regression/
        # That article assumes y in {-1, +1} whereas we use y in {0, 1}. Thus
        # the difference in formulas.
        # The same approach is used in sklearn.linear_model.LogisticRegression
        # and in LIBLINEAR
        if eta_out[i] > 0:
            expnegeta = exp(-eta_out[i])
            unit_deviance = weights[i] * (y[i] * eta_out[i] - eta_out[i] - log(1 + expnegeta))
            mu_out[i] = 1 / (1 + expnegeta)
        else:
            expposeta = exp(eta_out[i])
            unit_deviance = weights[i] * (y[i] * eta_out[i] - log(1 + expposeta))
            mu_out[i] = expposeta / (expposeta + 1)
        deviance += unit_deviance
    return -2 * deviance

def binomial_logit_rowwise_gradient_hessian(
    const_floating1d y,
    const_floating1d weights,
    const_floating1d eta,
    const_floating1d mu,
    floating[:] gradient_rows_out,
    floating[:] hessian_rows_out
):
    cdef int n = eta.shape[0]
    cdef int i
    for i in prange(n, nogil=True):
        gradient_rows_out[i] = weights[i] * (y[i] - mu[i])
        hessian_rows_out[i] = weights[i] * mu[i] * (1 - mu[i])



import time
from typing import Any, Callable, Dict, List, Tuple

import numpy as np
import pandas as pd
from tabmat.ext.dense import dense_sandwich


def _numpy_mklC(X, d):
    sqrtD = np.sqrt(d)[:, np.newaxis]
    x_d = X[0] * sqrtD
    return x_d.T @ x_d


def _numpy_mklF(X, d):
    sqrtD = np.sqrt(d)[:, np.newaxis]
    x_d = X[1] * sqrtD
    return x_d.T @ x_d


def _bench(f: Callable, iter: int) -> Tuple[List[float], Any]:
    ts = []
    for _ in range(iter):
        start = time.time()
        out: Any = f()
        ts.append(time.time() - start)
    return ts, out


def _dense_sandwichC(X, d):
    return dense_sandwich(X[0], d)


def _dense_sandwichF(X, d):
    return dense_sandwich(X[1], d)


def _mn_run(m, n, iter, dtype):
    precision = dtype().itemsize * 8
    X = [np.random.rand(n, m).astype(dtype=dtype)]
    d = np.random.rand(n).astype(dtype=dtype)

    X.append(np.asfortranarray(X[0]))

    out: Dict[str, Any] = {"name": [], "runtime": []}
    to_run = [
        "numpy_mklC",
        # "numpy_mklF",
        "_dense_sandwichC",
        "_dense_sandwichF",
    ]
    for name in to_run:
        ts, result = _bench(lambda: globals()[name](X, d), iter)
        if name == "numpy_mklC":
            true = result
        elif "numpy_mklC" in to_run:
            err = np.abs((true - result) / true)
            np.testing.assert_almost_equal(err, 0, 4 if precision == 32 else 7)
        runtime = np.min(ts)
        out["name"].append(name)
        out["runtime"].append(runtime)
        print(name, runtime)
    out_df = pd.DataFrame(out)
    out_df["m"] = m
    out_df["n"] = n
    out_df["precision"] = precision
    return out_df


def main():
    """Run some kind of benchmark."""
    iter = 20
    Rs = []
    for m, n in [
        (20, 1000000),
        # (50, 500000),
        # (150, 200000),
        # (300, 100000),
        # (2048, 2048),
        # (1500, 1500),
        (500, 500),
    ]:
        for dt in [np.float64]:
            Rs.append(_mn_run(m, n, iter, dt))
    df = pd.concat(Rs)
    df.set_index(["m", "n", "name", "precision"], inplace=True)
    df.sort_index(inplace=True)
    print(df)


def main2():
    """Run some kind of benchmark."""
    n = 500
    m = 500
    dtype = np.float64
    X = np.asfortranarray(np.random.rand(n, m).astype(dtype=dtype))
    d = np.random.rand(n).astype(dtype=dtype)
    t1d = []
    pls = []
    krs = []
    ibs = []
    results = []
    # for thresh1d in [16, 32, 64, 128]:
    #     for parlevel in [5, 7, 10, 13]:
    #         for kratio in [1, 10, 20, 80]:
    for thresh1d in [32, 64]:
        for parlevel in [9]:
            for kratio in [8, 16]:
                for innerblock in [32, 64, 128, 256]:
                    t1d.append(thresh1d)
                    pls.append(parlevel)
                    krs.append(kratio)
                    ibs.append(innerblock)
                    # results.append(np.min(bench(lambda: X.T @ X, 1)[0]))
                    results.append(
                        np.min(
                            _bench(
                                lambda: dense_sandwich(
                                    X, d, thresh1d, parlevel, kratio, innerblock
                                ),
                                50,
                            )[0]
                        )
                    )
                    print(results[-1])
    df = pd.DataFrame(
        dict(thresh1d=t1d, parlevel=pls, kratio=krs, innerblock=ibs, results=results)
    )
    df.set_index(["thresh1d", "parlevel", "kratio", "innerblock"], inplace=True)
    df.sort_index(inplace=True)
    with pd.option_context("display.max_rows", None, "display.max_columns", None):
        print(df)


#  841650213      L1-dcache-load-misses     #   12.01% of all L1-dcache hits    (71.87%)
# 7006517280      L1-dcache-loads                                               (71.53%)
# 1016757397      L1-dcache-stores                                              (69.82%)


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from glum_benchmarks.cli_run import get_all_problems
from glum_benchmarks.problems import Problem
from glum_benchmarks.util import (
    BenchmarkParams,
    exposure_and_offset_to_weights,
    get_obj_val,
    get_tweedie_p,
)

all_test_problems_offset = {
    k: v
    for k, v in get_all_problems().items()
    if "offset" in k and "gaussian" not in k and "binomial" not in k
}
bench_cfg = dict(
    num_rows=10000,
    regularization_strength=0.1,
    storage="dense",
)


@pytest.mark.parametrize(
    ["Pn", "P"],
    [
        x if "wide" not in x[0] else pytest.param(x[0], x[1], marks=pytest.mark.slow)
        for x in all_test_problems_offset.items()
    ],  # mark the "wide" problems as "slow" so that we can call pytest -m "not slow"
    ids=all_test_problems_offset.keys(),
)
def test_offset_solution_matches_weights_solution(
    Pn: str,
    P: Problem,
):
    params = BenchmarkParams(
        problem_name=Pn,
        library_name="sklearn-fork",
        # storage=storage,
        **bench_cfg,
    )

    tweedie_p = get_tweedie_p(P.distribution)

    dat = P.data_loader(num_rows=params.num_rows)
    weights_dat = {"X": dat["X"]}
    weights_dat["y"], weights_dat["weights"] = exposure_and_offset_to_weights(
        tweedie_p, dat["y"], offset=dat["offset"]
    )

    np.random.seed(0)
    coefs = np.zeros(dat["X"].shape[1] + 1)
    coefs[1:] = np.random.normal(0, 0.01 / dat["X"].std(0))
    eta = dat["X"].dot(coefs[1:]) + dat["offset"]
    coefs[0] = np.log(dat["y"].mean() / np.exp(eta).mean())

    eps = 1e-7
    coefs_2 = coefs + eps

    def get_obj_val_(dat, coefs):
        if "weights" in dat.keys():
            reg_multiplier = dat["weights"].mean()
        else:
            reg_multiplier = 1

        res = get_obj_val(
            dat,
            P.distribution,
            P.regularization_strength / reg_multiplier,
            P.l1_ratio,
            coefs[0],
            coefs[1:],
        )
        if "weights" in dat.keys():
            res *= reg_multiplier
        return res

    offset_result = get_obj_val_(dat, coefs)
    offset_result_2 = get_obj_val_(dat, coefs_2)
    weights_result = get_obj_val_(weights_dat, coefs)
    weights_result_2 = get_obj_val_(weights_dat, coefs_2)

    np.testing.assert_allclose(
        offset_result_2 - offset_result,
        weights_result_2 - weights_result,
        rtol=1e-7,
        atol=1e-14,
    )



import json
import warnings

import click
import numpy as np
import pytest
from git_root import git_root

from glum_benchmarks.cli_run import execute_problem_library
from glum_benchmarks.problems import Problem, get_all_problems
from glum_benchmarks.util import BenchmarkParams, get_obj_val

bench_cfg = dict(num_rows=10000, regularization_strength=0.1, diagnostics_level="none")

all_test_problems = get_all_problems()


def is_weights_problem_with_offset_match(problem_name):
    return (
        "no-weights" not in problem_name
        and "weights" in problem_name
        and (
            "gamma" in problem_name
            or "poisson" in problem_name
            or "tweedie" in problem_name
        )
    )


@pytest.fixture(scope="module")
def expected_all():
    with open(git_root("tests/glm/golden_master/benchmark_gm.json"), "r") as fh:
        return json.load(fh)


@pytest.mark.parametrize(
    ["Pn", "P"],
    [
        x if "wide" not in x[0] else pytest.param(x[0], x[1], marks=pytest.mark.slow)
        for x in all_test_problems.items()
    ],  # mark the "wide" problems as "slow" so that we can call pytest -m "not slow"
    ids=all_test_problems.keys(),
)
def test_gm_benchmarks(Pn: str, P: Problem, expected_all: dict):
    result, params = exec(Pn)

    if is_weights_problem_with_offset_match(Pn):
        expected = expected_all["offset".join(Pn.split("weights"))]
    else:
        expected = expected_all[Pn]

    all_result = np.concatenate(([result["intercept"]], result["coef"]))
    all_expected = np.concatenate(([expected["intercept"]], expected["coef"]))

    try:
        np.testing.assert_allclose(all_result, all_expected, rtol=2e-4, atol=2e-4)
    except AssertionError as e:
        dat = P.data_loader(
            num_rows=params.num_rows,
        )
        obj_result = get_obj_val(
            dat,
            P.distribution,
            P.regularization_strength,
            P.l1_ratio,
            all_result[0],
            all_result[1:],
        )
        expected_result = get_obj_val(
            dat,
            P.distribution,
            P.regularization_strength,
            P.l1_ratio,
            all_expected[0],
            all_expected[1:],
        )
        raise AssertionError(
            f"""Failed with error {e} on problem {Pn}.
            New objective function value is higher by {obj_result - expected_result}."""
        )


@click.command()
@click.option("--overwrite", is_flag=True, help="overwrite existing golden master")
@click.option(
    "--problem_name", default=None, help="Only run and store a specific problem."
)
def run_and_store_golden_master(overwrite, problem_name):

    try:
        with open(git_root("tests/glm/golden_master/benchmark_gm.json"), "r") as fh:
            gm_dict = json.load(fh)
    except FileNotFoundError:
        gm_dict = {}

    for Pn in get_all_problems().keys():
        if is_weights_problem_with_offset_match(Pn):
            continue
        if problem_name is not None:
            if Pn != problem_name:
                continue

        res, params = exec(Pn)

        if Pn in gm_dict.keys():
            if overwrite:
                warnings.warn("Overwriting existing result")
            else:
                warnings.warn("Result exists and cannot overwrite. Skipping")
                continue

        gm_dict[Pn] = dict(
            coef=res["coef"].tolist(),
            intercept=res["intercept"],
        )

    with open(git_root("tests/glm/golden_master/benchmark_gm.json"), "w") as fh:
        json.dump(gm_dict, fh, indent=2)


def exec(Pn):
    execute_args = ["diagnostics_level"]
    params = BenchmarkParams(
        problem_name=Pn,
        library_name="glum",
        **{k: v for k, v in bench_cfg.items() if k not in execute_args},
    )
    if bench_cfg["diagnostics_level"] != "none":
        print("Running", Pn)

    result, _ = execute_problem_library(
        params, **{k: v for k, v in bench_cfg.items() if k in execute_args}
    )
    return result, params


if __name__ == "__main__":
    run_and_store_golden_master()



import os
import sys

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

if sys.platform == "win32":
    allocator_libs = []  # type: ignore
    extra_compile_args = ["/openmp", "/O2"]
    extra_link_args = ["/openmp"]
else:
    allocator_libs = ["jemalloc"]
    extra_compile_args = [
        "-fopenmp",
        "-O3",
        "-ffast-math",
        "--std=c++17",
    ]
    extra_link_args = ["-fopenmp"]

architecture = os.environ.get("GLM_ARCHITECTURE", "native")
if architecture != "default":
    extra_compile_args.append("-march=" + architecture)

extension_args = dict(
    include_dirs=[np.get_include()],
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c++",
)
ext_modules = [
    Extension(
        name="glum._functions",
        sources=["src/glum/_functions.pyx"],
        **extension_args,
    ),
    Extension(
        name="glum._cd_fast",
        sources=["src/glum/_cd_fast.pyx"],
        **extension_args,
    ),
]

setup(
    name="glum",
    use_scm_version={"version_scheme": "post-release"},
    setup_requires=["setuptools_scm"],
    description="High performance Python GLMs with all the features!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Quantco/glum",
    author="QuantCo, Inc.",
    author_email="noreply@quantco.com",
    license="BSD",
    classifiers=[  # Optional
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    package_dir={"": "src"},
    packages=find_packages(
        where="src",
        include=["glum"]
        if os.environ.get("CONDA_BUILD")
        else ["glum", "glum_benchmarks"],
    ),
    install_requires=[
        "joblib",
        "numexpr",
        "numpy",
        "pandas",
        "scikit-learn>=0.23",
        "scipy",
        "tabmat>=3.0.1",
    ],
    entry_points=None
    if os.environ.get("CONDA_BUILD")
    else """
        [console_scripts]
        glm_benchmarks_run = glum_benchmarks.cli_run:cli_run
        glm_benchmarks_analyze = glum_benchmarks.cli_analyze:cli_analyze
    """,
    ext_modules=cythonize(ext_modules, annotate=False),
    zip_safe=False,
)



import gc
import multiprocessing as mp
import time
import warnings
from threading import Thread
from typing import List, Optional

import numpy as np
import pandas as pd
import psutil
import scipy.sparse as sps
import tabmat as tm

from glum import GeneralizedLinearRegressor
from glum_benchmarks.cli_run import get_all_problems
from glum_benchmarks.util import get_sklearn_family, runtime


def _get_memory_usage() -> int:
    return psutil.Process().memory_info().rss


class MemoryPoller:
    """
    Sample memory and compute useful statistics.

    Example usage:

    with MemoryPoller() as mp:
        do some stuff here
        print('initial memory usage', mp.initial_memory)
        print('max memory usage', mp.max_memory)
        excess_memory_used = mp.max_memory - mp.initial_memory
    """

    def _poll_max_memory_usage(self):
        while not self.stop_polling:
            self.memory_usage.append(_get_memory_usage())
            self.max_memory: int = max(self.max_memory, self.memory_usage[-1])
            time.sleep(1e-4)

    def __enter__(self):
        """See example usage above."""
        self.stop_polling = False
        self.max_memory = 0
        self.initial_memory = _get_memory_usage()
        self.memory_usage: List[int] = [self.initial_memory]
        self.t = Thread(target=self._poll_max_memory_usage)
        self.t.start()
        return self

    def __exit__(self, *excargs):
        """Stop polling memory usage."""
        self.stop_polling = True
        self.t.join()


def _get_x_bytes(x) -> int:
    if isinstance(x, np.ndarray):
        return x.nbytes
    if sps.issparse(x):
        return sum(
            mat.data.nbytes + mat.indices.nbytes + mat.indptr.nbytes
            for mat in [x, x.x_csr]
        )
    if isinstance(x, tm.CategoricalMatrix):
        return x.indices.nbytes
    if isinstance(x, tm.SplitMatrix):
        return sum(_get_x_bytes(elt) for elt in x.matrices)
    raise NotImplementedError(f"Can't get bytes for matrix of type {type(x)}.")


def _runner(storage, copy_X: Optional[bool]):
    gc.collect()

    P = get_all_problems()["wide-insurance-no-weights-lasso-poisson"]
    dat = P.data_loader(num_rows=100000, storage=storage)

    # Measure how much memory we are using before calling the GLM code
    if isinstance(dat["X"], pd.DataFrame):
        X = dat["X"].to_numpy()
    elif sps.issparse(dat["X"]):
        X = tm.SparseMatrix(dat["X"])
    elif isinstance(dat["X"], tm.SplitMatrix):
        X = dat["X"]

    data_memory = _get_x_bytes(X)

    y = dat["y"]
    del dat
    gc.collect()

    with MemoryPoller() as mp:
        for _ in range(4):
            model = GeneralizedLinearRegressor(
                family="poisson",
                l1_ratio=1.0,
                alpha=0.01,
                force_all_finite=False,
                copy_X=copy_X,
            )
            model.fit(X=X, y=y)

        excess_memory_used = mp.max_memory - mp.initial_memory
        extra_to_initial_ratio = excess_memory_used / data_memory

        import matplotlib.pyplot as plt

        graph = (np.array(mp.memory_usage) - mp.initial_memory) / data_memory
        plt.plot(graph)
        plt.ylabel("Memory (fraction of X)")
        plt.xlabel("Time (1e-4s)")
        plt.savefig(f"performance/memory_{storage}_copy_{copy_X}.png")

        return extra_to_initial_ratio


def _make_memory_usage_plots():
    # These values are around double the empirical extra memory used. They inc
    # They increase from dense->sparse->split->cat, because the matrix itself takes less
    # and less memory to store, so all the temporary vectors of length n_rows start to
    # dominate the memory usage.
    storage_allowed_ratio = {"dense": 0.1, "sparse": 0.45, "cat": 1.3, "split0.1": 0.55}
    for storage, allowed_ratio in storage_allowed_ratio.items():
        for copy_X in [False, True, None]:
            with mp.Pool(1) as p:
                extra_to_initial_ratio = p.starmap(_runner, [(storage, copy_X)])[0]

            if copy_X is not None and copy_X:
                if extra_to_initial_ratio < 1:
                    warnings.warn(
                        f"Used less memory than expected with copy_X = True and "
                        f"data format {storage}. Memory exceeded initial memory by "
                        f"{extra_to_initial_ratio}."
                    )
            else:
                if extra_to_initial_ratio > allowed_ratio:
                    warnings.warn(
                        f"Used more memory than expected with copy_X = {copy_X} and "
                        f"data format {storage}. Memory exceeded initial memory by "
                        f"{extra_to_initial_ratio}; expected less than {allowed_ratio}."
                    )


def get_spmv_runtime():
    """
    Get runtime of sparse matrix-vector product.

    Sparse matrix-vector product runtime should be representative of the memory
    bandwidth of the machine. Automatically scale the according to half the
    number of cores since the scipy.sparse implementation is not parallelized
    and glum is parallelized.
    """
    N = 20000000
    diag_data = np.random.rand(5, N)
    mat = sps.spdiags(diag_data, [0, 1, -1, 2, -2], N, N).tocsr()
    v = np.random.rand(N)
    return runtime(lambda: mat.dot(v), 5)[0] / (mp.cpu_count() // 2)


def get_dense_inv_runtime():
    """
    Get runtime of dense matrix inverse.

    Dense matrix multiplication runtime should be representative of the
    floating point performance of the machine.
    """
    N = 1300
    X = np.random.rand(N, N)
    return runtime(lambda: np.linalg.inv(X), 5)[0]


def runtime_checker():
    """
    Run various operations and check that glum doesn't run too much
    slower than operations expected to be similar. This isn't a perfect test
    but it'll raise a red flag if the code has unexpectedly gotten much slower.
    """
    spmv_runtime = get_spmv_runtime()
    dense_inv_runtime = get_dense_inv_runtime()

    what_to_check = [
        ("dense", "narrow-insurance-no-weights-lasso", "poisson", 200000, 1.5),
        ("sparse", "narrow-insurance-weights-l2", "gaussian", 200000, 2.5),
        ("cat", "wide-insurance-no-weights-l2", "gamma", 100000, 2.5),
        ("split0.1", "wide-insurance-offset-lasso", "tweedie-p=1.5", 100000, 3.0),
        ("split0.1", "intermediate-insurance-no-weights-net", "binomial", 200000, 1.0),
    ]

    for storage, problem, distribution, num_rows, limit in what_to_check:
        P = get_all_problems()[problem + "-" + distribution]
        dat = P.data_loader(num_rows=num_rows, storage=storage)

        family = get_sklearn_family(distribution)
        model = GeneralizedLinearRegressor(
            family=family,
            l1_ratio=1.0,
            alpha=0.01,
            copy_X=False,
            force_all_finite=False,
        )
        min_runtime, result = runtime(lambda: model.fit(X=dat["X"], y=dat["y"]), 5)

        # Let's just guess that we're about half flop-limited and half
        # memory-limited.  This is a decent guess because the sandwich product is
        # mostly flop-limited in the dense case and the dense case generally
        # dominates even when we're using split or categorical. On the other hand,
        # everything besides the sandwich product is probably memory limited.
        denominator = 0.5 * dense_inv_runtime + 0.5 * spmv_runtime
        if min_runtime / denominator > limit:
            warnings.warn(
                f"runtime {min_runtime} is greater than the expected maximum runtime "
                f"of {limit * denominator}"
            )


if __name__ == "__main__":
    # make_memory_usage_plots()
    runtime_checker()



import os
import tempfile
import warnings
from typing import Any, Dict, List

import click
import pytest
from click.testing import CliRunner

from glum_benchmarks.cli_analyze import _identify_parameter_fnames
from glum_benchmarks.cli_run import cli_run
from glum_benchmarks.util import BenchmarkParams, benchmark_params_cli, defaults


@pytest.mark.parametrize(
    "cli_options, expected_params",
    [
        ([], {}),
        (["--num_rows", "1000", "--cv", "True"], {"num_rows": 1000, "cv": True}),
    ],
)
def test_make_params(cli_options: List[str], expected_params: Dict[str, Any]):
    """
    Test that the basic command line interface runs and that the benchmark_params_cli \
    decorator works.

    Parameters
    ----------
    cli_options: List of strings
    expected_params
    """

    @click.command()
    @benchmark_params_cli
    def _params_test(params: BenchmarkParams):
        for k in params.param_names:
            as_expected = getattr(params, k) == expected_params.get(k)
            if not as_expected:
                click.echo(
                    f"""
                    For parameter {k} expected {expected_params.get(k)} but
                    got {getattr(params, k)}."""
                )

    runner = CliRunner()
    result = runner.invoke(_params_test, cli_options)
    if not result.exit_code == 0:
        raise ValueError(result.output)


def test_correct_problems_run():
    """Test that the correct problems are run given certain command-line inputs."""
    output_dir = "test_output_tmp"

    problem_names = [
        "narrow-insurance-weights-l2-gamma",
        "wide-insurance-no-weights-net-poisson",
    ]
    library_names = ["zeros", "glum"]
    num_rows = 20
    regularization_strength = 1000.0

    assert output_dir not in os.listdir()
    with tempfile.TemporaryDirectory() as d:
        args = [
            "--problem_name",
            ",".join(problem_names),
            "--library_name",
            ",".join(library_names),
            "--num_rows",
            str(num_rows),
            "--regularization_strength",
            str(regularization_strength),
            "--output_dir",
            d,
        ]
        runner = CliRunner()
        result = runner.invoke(cli_run, args)
        if not result.exit_code == 0:
            problem_name_str = " ".join(args)
            raise ValueError(
                f"""Failed on problem {problem_name_str} with output: \n {result.output}"""
            )
        problems_run = os.listdir(d)

    expected_problems_run = [
        BenchmarkParams(
            pn,
            ln,
            num_rows=num_rows,
            regularization_strength=regularization_strength,
            **{
                k: defaults[k]
                for k in [
                    "storage",
                    "threads",
                    "single_precision",
                    "cv",
                    "hessian_approx",
                    "diagnostics_level",
                ]
            },
        ).get_result_fname()
        + ".pkl"
        for pn in problem_names
        for ln in library_names
    ]

    n_threads = os.environ.get("OMP_NUM_THREADS", os.cpu_count())

    expected_problems_run_2 = [
        f"narrow-insurance-weights-l2-gamma_zeros_20_dense_{n_threads}_False_1000.0_Fals"
        "e_0.0_basic.pkl",
        f"narrow-insurance-weights-l2-gamma_glum_20_dense_{n_threads}_False_10"
        "00.0_False_0.0_basic.pkl",
        f"wide-insurance-no-weights-net-poisson_zeros_20_dense_{n_threads}_False_1000.0"
        "_False_0.0_basic.pkl",
        f"wide-insurance-no-weights-net-poisson_glum_20_dense_{n_threads}_False"
        "_1000.0_False_0.0_basic.pkl",
    ]

    assert sorted(problems_run) == sorted(expected_problems_run)
    assert sorted(problems_run) == sorted(expected_problems_run_2)


def test_correct_problems_analyzed():
    """
    Test that cli_analyze runs on the correct problems.

    Everything should be analyzed when cli_analyze is not given any parameters.
    This test checks that everything in benchmark_output/ is run, so if that directory
    is empty, the test will not be meaningful.
    """
    output_dir = "benchmark_output"
    if output_dir not in os.listdir():
        warnings.warn("Output directory not found")
        return

    to_analyze = _identify_parameter_fnames(output_dir, BenchmarkParams())
    assert sorted(to_analyze) == sorted(os.listdir(output_dir))



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
import subprocess

import matplotlib.pyplot as plt
import pandas as pd

# %%
# !rm -r benchmark_output

# %%
base_cmd = (
    "glm_benchmarks_run --threads 6 --num_rows {n} --storage {s} "
    "--problem_name {p} --library_name {lib}"
)

problems = []
for p in ["narrow-insurance", "intermediate-insurance", "wide-insurance"]:
    for reg in ["l2", "lasso"]:
        for dist in ["tweedie-p=1.5", "poisson", "gaussian", "gamma", "binomial"]:
            problems.append(f"{p}-no-weights-{reg}-{dist}")

p = "intermediate-housing"
for reg in ["l2", "lasso"]:
    for dist in ["gaussian", "gamma", "binomial"]:
        problems.append(f"{p}-no-weights-{reg}-{dist}")

n = 500000

# %%
# run r-glmnet and h2o benchmarks, sparse storage works best.
s = "sparse"
for lib in ["r-glmnet", "h2o"]:
    for p in problems:
        if lib == "r-glmnet" and p == "wide-insurance-no-weights-l2-poisson":
            continue
        cmd = base_cmd.format(n=n, s=s, p=p, lib=lib)
        print(cmd)
        subprocess.run(cmd.split(" "))

# run glum benchmarks where auto storage works best.
lib = "glum"
s = "auto"
for p in problems:
    cmd = base_cmd.format(n=n, s=s, p=p, lib=lib)
    print(cmd)
    subprocess.run(cmd.split(" "))

analyze_cmd = "glm_benchmarks_analyze --export benchmark_data.csv"
subprocess.run(analyze_cmd.split(" "))

# %%
df = pd.read_csv("benchmark_data.csv")
df.drop(
    [
        "storage",
        "num_rows",
        "regularization_strength",
        "offset",
        "threads",
        "single_precision",
        "cv",
        "hessian_approx",
        "diagnostics_level",
    ],
    axis=1,
    inplace=True,
)
df["distribution"] = (
    df["problem_name"].str.split("-").apply(lambda x: x[-2] if "5" in x[-1] else x[-1])
)

# %%
# %config InlineBackend.figure_format='retina'

# %%
for prob_name in ["narrow-insurance", "intermediate-insurance", "intermediate-housing"]:
    for reg in ["l2", "lasso"]:
        plot_df = (
            df[
                df["problem_name"].str.contains(reg)
                & df["problem_name"].str.contains(prob_name)
            ]
            .copy()
            .set_index(["distribution"])[["runtime", "library_name"]]
        )
        plot_df = plot_df.pivot(columns="library_name")
        plot_df.columns = plot_df.columns.get_level_values(1)
        plot_df = plot_df.sort_index(axis=1).rename(columns={"r-glmnet": "glmnet"})
        plot_df.index = [x.title() for x in plot_df.index]

        title = prob_name.title() + "-" + ("Lasso" if reg == "lasso" else "Ridge")
        plot_df.plot.bar(
            ylim=[0, 4],
            title=title,
            legend=False,
            figsize=(6, 3),
            width=0.8,
            ylabel="run time (s)",
            yticks=[0, 1, 2, 3, 4],
            cmap="Paired",
        )
        plt.legend(bbox_to_anchor=(1, 1), loc="upper left", ncol=1)
        plt.xticks(rotation=45, ha="right")

        ax = plt.gca()

        # Hide the right and top spines
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)

        # Only show ticks on the left and bottom spines
        ax.yaxis.set_ticks_position("left")
        ax.xaxis.set_ticks_position("bottom")

        for p in ax.patches:
            x = p.get_x()  # type: ignore
            y = p.get_height()  # type: ignore
            if y > 4.1:
                text_x = x + 0.02
                text_y = 2.75 if y > 10 else 2.95
                ax.annotate(
                    f"{y:.1f}",
                    (text_x, text_y),
                    fontsize=14,
                    rotation="vertical",
                )
                arrow_x = text_x + 0.11
                arrow_y = 3.5
                ax.annotate(
                    "",
                    xy=(arrow_x, arrow_y + 0.5),
                    xytext=(arrow_x, arrow_y),
                    arrowprops=dict(arrowstyle="->"),
                )

        plt.tight_layout()
        fp = f"../_static/{prob_name}-{reg}.png"
        plt.savefig(fp, dpi=300)
        fp = f"../_static/{prob_name}-{reg}.pdf"
        plt.savefig(fp)
        plt.show()

# %%
for prob_name in ["wide-insurance"]:
    for reg in ["l2", "lasso"]:
        plot_df = (
            df[
                df["problem_name"].str.contains(reg)
                & df["problem_name"].str.contains(prob_name)
            ]
            .copy()
            .set_index(["distribution"])[["runtime", "library_name"]]
        )
        plot_df = plot_df.pivot(columns="library_name")
        plot_df.columns = plot_df.columns.get_level_values(1)
        plot_df = plot_df.sort_index(axis=1).rename(columns={"r-glmnet": "glmnet"})
        plot_df.index = [x.title() for x in plot_df.index]

        title = prob_name.title() + "-" + ("Lasso" if reg == "lasso" else "Ridge")
        plot_df.plot.bar(
            ylim=[0, 15],
            title=title,
            legend=False,
            figsize=(6, 3),
            width=0.8,
            ylabel="run time (s)",
            yticks=[0, 3, 6, 9, 12, 15],
            cmap="Paired",
        )
        plt.legend(bbox_to_anchor=(1, 1), loc="upper left", ncol=1)
        plt.xticks(rotation=45, ha="right")

        ax = plt.gca()

        # Hide the right and top spines
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)

        # Only show ticks on the left and bottom spines
        ax.yaxis.set_ticks_position("left")
        ax.xaxis.set_ticks_position("bottom")

        for p in ax.patches:
            x = p.get_x()  # type: ignore
            y = p.get_height()  # type: ignore
            if y > 15.1:
                text_x = x + 0.01
                text_y = 8 if y > 1000 else (9 if y > 100 else 10)
                ax.annotate(
                    f"{y:.1f}",
                    (text_x, text_y),
                    fontsize=14,
                    rotation="vertical",
                )
                arrow_x = text_x + 0.13
                arrow_y = 13
                ax.annotate(
                    "",
                    xy=(arrow_x, arrow_y + 2),
                    xytext=(arrow_x, arrow_y),
                    arrowprops=dict(arrowstyle="->"),
                )

        plt.tight_layout()
        fp = f"../_static/{prob_name}-{reg}.png"
        plt.savefig(fp, dpi=300)
        fp = f"../_static/{prob_name}-{reg}.pdf"
        plt.savefig(fp)
        plt.show()

# %%
prob_name = "intermediate-insurance"
reg = "lasso"
plot_df = (
    df[
        df["problem_name"].str.contains(reg)
        & df["problem_name"].str.contains(prob_name)
    ]
    .copy()
    .set_index(["distribution"])[["runtime", "library_name"]]
)
plot_df = plot_df.pivot(columns="library_name")
plot_df.columns = plot_df.columns.get_level_values(1)
plot_df = plot_df.sort_index(axis=1).rename(columns={"r-glmnet": "glmnet"})
plot_df.index = [x.title() for x in plot_df.index]

plot_df.plot.bar(
    ylim=[0, 5],
    legend=False,
    figsize=(6, 3),
    width=0.8,
    ylabel="run time (s)",
    yticks=[0, 1, 2, 3, 4, 5],
    cmap="Paired",
)
plt.legend(bbox_to_anchor=(0, 1), loc="upper left", ncol=1)
plt.xticks(rotation=45, ha="right")

ax = plt.gca()

# Hide the right and top spines
ax.spines["right"].set_visible(False)
ax.spines["top"].set_visible(False)

# Only show ticks on the left and bottom spines
ax.yaxis.set_ticks_position("left")
ax.xaxis.set_ticks_position("bottom")

for p in ax.patches:
    x = p.get_x()  # type: ignore
    y = p.get_height()  # type: ignore
    if y > 5.1:
        text_x = x + 0.04
        text_y = 3.2
        ax.annotate(
            f"{y:.1f}",
            (text_x, text_y),
            fontsize=14,
            rotation="vertical",
        )
        arrow_x = text_x + 0.15
        arrow_y = 4.2
        ax.annotate(
            "",
            xy=(arrow_x, arrow_y + 0.8),
            xytext=(arrow_x, arrow_y),
            arrowprops=dict(arrowstyle="->"),
        )

plt.tight_layout()
fp = "../_static/headline_benchmark.png"
plt.savefig(fp, dpi=300)
fp = "../_static/headline_benchmark.pdf"
plt.savefig(fp)
plt.show()



import warnings
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
from scipy import sparse as sps

from .orig_sklearn_fork import (  # type: ignore
    GeneralizedLinearRegressor,
    TweedieDistribution,
)
from .util import benchmark_convergence_tolerance, runtime

random_seed = 110


def _build_and_fit(model_args, fit_args):
    return GeneralizedLinearRegressor(**model_args).fit(**fit_args)


def orig_sklearn_fork_bench(
    dat: Dict[str, Union[np.ndarray, sps.spmatrix]],
    distribution: str,
    alpha: float,
    l1_ratio: float,
    iterations: int,
    cv: bool,
    reg_multiplier: Optional[float] = True,
    **kwargs,
):
    """
    Benchmark the original sklearn-fork.

    Parameters
    ----------
    dat
    distribution
    alpha
    l1_ratio
    iterations
    cv
    reg_multiplier
    kwargs

    Returns
    -------
    Dict of
    """
    if cv:
        raise ValueError("original sklearn fork does not support cross-validation")
    result: Dict[str, Any] = {}

    X = dat["X"]

    if X.shape[0] > 100000 and not isinstance(X, (np.ndarray, pd.DataFrame)):
        warnings.warn(
            "original sklearn fork is too slow on sparse data sets with more than "
            "100,000 rows. Skipping."
        )
        return result

    fit_args = dict(X=X, y=dat["y"])
    if "sample_weight" in dat.keys():
        fit_args.update({"sample_weight": dat["sample_weight"]})
    if "offset" in dat.keys():
        warnings.warn("Original sklearn_fork does not support offsets.")
        return result

    family = distribution
    if family == "gaussian":
        family = "normal"
    elif "tweedie" in family:
        tweedie_p = float(family.split("-p=")[1])
        family = TweedieDistribution(tweedie_p)  # type: ignore

    model_args = dict(
        family=family,
        alpha=alpha if reg_multiplier is None else alpha * reg_multiplier,
        l1_ratio=l1_ratio,
        max_iter=150,
        random_state=random_seed,
        copy_X=False,
        selection="cyclic",
        tol=benchmark_convergence_tolerance,
    )

    try:
        result["runtime"], m = runtime(_build_and_fit, iterations, model_args, fit_args)
    except ValueError as e:
        warnings.warn(f"Problem failed with this error: {e}")
        return result

    result["intercept"] = m.intercept_
    result["coef"] = m.coef_
    result["n_iter"] = m.n_iter_

    return result



import warnings
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
import rpy2.robjects as ro
import rpy2.robjects.numpy2ri as n2r
from rpy2.robjects.packages import importr
from scipy import sparse as sps

from .util import benchmark_convergence_tolerance, runtime

is_initialized = False


def _r_install_if_not_available(pkg_name):
    utils = importr("utils")
    if pkg_name not in utils.installed_packages():
        utils.install_packages(pkg_name, repos="https://cloud.r-project.org")


def _setup_r_glmnet():
    global is_initialized
    if is_initialized:
        return
    n2r.activate()

    ro.r.library("utils")
    for pkg in ["glmnet", "statmod", "tweedie"]:
        _r_install_if_not_available(pkg)
        ro.r.library(pkg)

    is_initialized = True


def _to_r_obj(X, R_name):
    nr = X.shape[0]
    nc = 1 if len(X.shape) == 1 else X.shape[1]
    if sps.issparse(X):
        r_Matrix = importr("Matrix")
        X_coo = X.tocoo()
        ro.r.assign(
            R_name,
            r_Matrix.sparseMatrix(
                i=ro.IntVector(X_coo.row + 1),
                j=ro.IntVector(X_coo.col + 1),
                x=ro.FloatVector(X_coo.data),
                dims=ro.IntVector(X_coo.shape),
            ),
        )
    else:  # if dense
        ro.r.assign(R_name, ro.r.matrix(X, nrow=nr, ncol=nc))


def r_glmnet_bench(
    dat: Dict[str, Union[sps.spmatrix, np.ndarray]],
    distribution: str,
    alpha: float,
    l1_ratio: float,
    iterations: int,
    cv: bool,
    reg_multiplier: Optional[float] = None,
    **kwargs,
) -> Dict[str, Any]:
    """
    Run glmnet benchmark in R then port the results back to Python.

    Parameters
    ----------
    dat
    distribution
    alpha
    l1_ratio
    iterations
    cv
    reg_multiplier
    kwargs

    Returns
    -------
    Dict storing info about this model run.
    """
    result: Dict = {}

    _setup_r_glmnet()

    X = dat["X"]
    if isinstance(X, sps.spmatrix):
        if not isinstance(X, sps.csc.csc_matrix):
            warnings.warn("sparse matrix will be converted to csc format")
            X = X.tocsc()
    elif isinstance(X, pd.DataFrame):
        X = X.to_numpy()
    elif not isinstance(X, np.ndarray):
        warnings.warn(
            "glmnet requires data as scipy.sparse matrix, pandas dataframe, or "
            "numpy array. Skipping."
        )
        return result

    if distribution == "gamma":
        distribution = ro.r["Gamma"](link="log")
    elif distribution.startswith("tweedie"):
        p = float(distribution.split("tweedie-p=")[1])
        distribution = ro.r["tweedie"](link_power=0, var_power=p)

    r = ro.r

    # Do this before fitting so we're not including python to R conversion
    # times
    _to_r_obj(X, "X_in_R")
    _to_r_obj(dat["y"], "y_in_R")

    glmnet_kws = dict(
        x=r["X_in_R"],
        y=r["y_in_R"],
        family=distribution,
        alpha=l1_ratio,
        standardize=False,
        thresh=benchmark_convergence_tolerance,
    )
    if "sample_weight" in dat.keys():
        glmnet_kws.update({"weights": ro.FloatVector(dat["sample_weight"])})
    if "offset" in dat.keys():
        glmnet_kws.update({"offset": ro.FloatVector(dat["offset"])})

    # By default, glmnet runs for 100 different values of regularization strength.
    # For a fair comparison of runtime, we'd like to run for just a single value.
    # These parameters ensure that is the case.
    glmnet_kws["lambda"] = alpha
    glmnet_kws["nlambda"] = 1

    # NOTE: We checked thoroughly and this runtime measurement only includes
    # The cost of the glmnet function in R. We checked this by running the same
    # problem directly from R.
    result["runtime"], m = runtime(r["glmnet"], iterations, **glmnet_kws)
    result["intercept"] = m.rx2("a0")[0]
    result["coef"] = np.squeeze(np.asanyarray(r["as.matrix"](m.rx2("beta"))))
    result["n_iter"] = m.rx2("npasses")[0]
    return result
